@@ -5,9 +5,27 @@ from unittest.mock import MagicMock, call, patch
 import requests_mock
 
 from streamlink import Streamlink
-from streamlink.plugin import PluginError
-from streamlink.plugins.twitch import Twitch, TwitchHLSStream
-from tests.mixins.stream_hls import Playlist, Segment as _Segment, Tag, TestMixinStreamHLS
+from streamlink.plugins.twitch import Twitch, TwitchHLSStream, TwitchHLSStreamReader, TwitchHLSStreamWriter
+from tests.mixins.stream_hls import EventedHLSStreamWriter, Playlist, Segment as _Segment, Tag, TestMixinStreamHLS
+from tests.plugins import PluginCanHandleUrl
+
+
+class TestPluginCanHandleUrlTwitch(PluginCanHandleUrl):
+    __plugin__ = Twitch
+
+    should_match = [
+        'https://www.twitch.tv/twitch',
+        'https://www.twitch.tv/videos/150942279',
+        'https://clips.twitch.tv/ObservantBenevolentCarabeefPhilosoraptor',
+        'https://www.twitch.tv/weplaydota/clip/FurryIntelligentDonutAMPEnergyCherry-akPRxv7Y3w58WmFq'
+        'https://www.twitch.tv/twitch/video/292713971',
+        'https://www.twitch.tv/twitch/v/292713971',
+    ]
+
+    should_not_match = [
+        'https://www.twitch.tv',
+        'https://www.twitch.tv/',
+    ]
 
 
 DATETIME_BASE = datetime(2000, 1, 1, 0, 0, 0, 0)
@@ -44,29 +62,21 @@ class SegmentPrefetch(Segment):
         return "#EXT-X-TWITCH-PREFETCH:{0}".format(self.url(namespace))
 
 
-class TestPluginTwitch(unittest.TestCase):
-    def test_can_handle_url(self):
-        should_match = [
-            'https://www.twitch.tv/twitch',
-            'https://www.twitch.tv/videos/150942279',
-            'https://clips.twitch.tv/ObservantBenevolentCarabeefPhilosoraptor',
-            'https://www.twitch.tv/twitch/video/292713971',
-            'https://www.twitch.tv/twitch/v/292713971',
-        ]
-        for url in should_match:
-            self.assertTrue(Twitch.can_handle_url(url))
+class _TwitchHLSStreamWriter(EventedHLSStreamWriter, TwitchHLSStreamWriter):
+    pass
 
-    def test_can_handle_url_negative(self):
-        should_not_match = [
-            'https://www.twitch.tv',
-        ]
-        for url in should_not_match:
-            self.assertFalse(Twitch.can_handle_url(url))
+
+class _TwitchHLSStreamReader(TwitchHLSStreamReader):
+    __writer__ = _TwitchHLSStreamWriter
+
+
+class _TwitchHLSStream(TwitchHLSStream):
+    __reader__ = _TwitchHLSStreamReader
 
 
 @patch("streamlink.stream.hls.HLSStreamWorker.wait", MagicMock(return_value=True))
 class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
-    __stream__ = TwitchHLSStream
+    __stream__ = _TwitchHLSStream
 
     def get_session(self, options=None, disable_ads=False, low_latency=False):
         session = super().get_session(options)
@@ -82,6 +92,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(0, [daterange, Segment(0), Segment(1)], end=True)
         ], disable_ads=True, low_latency=False)
 
+        self.await_write(2)
         self.assertEqual(self.await_read(read_all=True), self.content(segments), "Doesn't filter out segments")
         self.assertTrue(all(self.called(s) for s in segments.values()), "Downloads all segments")
 
@@ -91,6 +102,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(0, [daterange, Segment(0), Segment(1)], end=True)
         ], disable_ads=True, low_latency=False)
 
+        self.await_write(2)
         self.assertEqual(self.await_read(read_all=True), segments[1].content, "Filters out ad segments")
         self.assertTrue(all(self.called(s) for s in segments.values()), "Downloads all segments")
 
@@ -100,6 +112,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(0, [daterange, Segment(0), Segment(1)], end=True)
         ], disable_ads=True, low_latency=False)
 
+        self.await_write(2)
         self.assertEqual(self.await_read(read_all=True), segments[1].content, "Filters out ad segments")
         self.assertTrue(all(self.called(s) for s in segments.values()), "Downloads all segments")
 
@@ -109,6 +122,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(0, [daterange, Segment(0), Segment(1)], end=True)
         ], disable_ads=True, low_latency=False)
 
+        self.await_write(2)
         self.assertEqual(self.await_read(read_all=True), segments[1].content, "Filters out ad segments")
         self.assertTrue(all(self.called(s) for s in segments.values()), "Downloads all segments")
 
@@ -121,6 +135,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(4, [Segment(4), Segment(5)], end=True)
         ], disable_ads=True, low_latency=False)
 
+        self.await_write(6)
         self.assertEqual(
             self.await_read(read_all=True),
             self.content(segments, cond=lambda s: s.num >= 4),
@@ -141,6 +156,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(4, [Segment(4), Segment(5)], end=True)
         ], disable_ads=True, low_latency=False)
 
+        self.await_write(6)
         self.assertEqual(
             self.await_read(read_all=True),
             self.content(segments, cond=lambda s: s.num != 2 and s.num != 3),
@@ -159,6 +175,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(2, [Segment(2), Segment(3)], end=True)
         ], disable_ads=False, low_latency=False)
 
+        self.await_write(4)
         self.assertEqual(
             self.await_read(read_all=True),
             self.content(segments),
@@ -175,8 +192,8 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
         ], disable_ads=False, low_latency=True)
 
         self.assertEqual(2, self.session.options.get("hls-live-edge"))
-        self.assertEqual(True, self.session.options.get("hls-segment-stream-data"))
 
+        self.await_write(6)
         self.assertEqual(
             self.await_read(read_all=True),
             self.content(segments, cond=lambda s: s.num >= 4),
@@ -196,8 +213,8 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
         ], disable_ads=False, low_latency=False)
 
         self.assertEqual(4, self.session.options.get("hls-live-edge"))
-        self.assertEqual(False, self.session.options.get("hls-segment-stream-data"))
 
+        self.await_write(8)
         self.assertEqual(
             self.await_read(read_all=True),
             self.content(segments, cond=lambda s: s.num < 8),
@@ -217,6 +234,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
         self.assertTrue(self.session.get_plugin_option("twitch", "low-latency"))
         self.assertFalse(self.session.get_plugin_option("twitch", "disable-ads"))
 
+        self.await_write(6)
         self.await_read(read_all=True)
         self.assertEqual(mock_log.info.mock_calls, [
             call("Low latency streaming (HLS live edge: 2)"),
@@ -231,6 +249,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7), SegmentPrefetch(8), SegmentPrefetch(9)], end=True)
         ], disable_ads=False, low_latency=True)
 
+        self.await_write(8)
         self.assertEqual(
             self.await_read(read_all=True),
             self.content(segments, cond=lambda s: s.num > 1),
@@ -250,6 +269,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7), SegmentPrefetch(8), SegmentPrefetch(9)], end=True)
         ], disable_ads=True, low_latency=True)
 
+        self.await_write(8)
         self.await_read(read_all=True)
         self.assertEqual(mock_log.info.mock_calls, [
             call("Will skip ad segments"),
@@ -265,6 +285,7 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
             Playlist(4, [Segment(4), Segment(5), Segment(6), Segment(7)], end=True)
         ], disable_ads=True, low_latency=True)
 
+        self.await_write(6)
         self.await_read(read_all=True)
         self.assertEqual(mock_log.info.mock_calls, [
             call("Will skip ad segments"),
@@ -277,100 +298,141 @@ class TestTwitchHLSStream(TestMixinStreamHLS, unittest.TestCase):
 class TestTwitchMetadata(unittest.TestCase):
     def setUp(self):
         self.mock = requests_mock.Mocker()
+        self.mock.register_uri(requests_mock.ANY, requests_mock.ANY, exc=requests_mock.exceptions.InvalidRequest)
         self.mock.start()
 
     def tearDown(self):
         self.mock.stop()
 
-    def subject(self, url):
+    @staticmethod
+    def subject(url):
         session = Streamlink()
         Twitch.bind(session, "tests.plugins.test_twitch")
         plugin = Twitch(url)
         return plugin.get_author(), plugin.get_title(), plugin.get_category()
 
-    def subject_channel(self, data=True, failure=False):
-        self.mock.get(
-            "https://api.twitch.tv/kraken/users?login=foo",
-            json={"users": [{"_id": 1234}]}
+    def mock_request_channel(self, data=True):
+        return self.mock.post(
+            "https://gql.twitch.tv/gql",
+            json=[
+                {"data": {"userOrError": {"userDoesNotExist": "error"} if not data else {
+                    "displayName": "channel name"
+                }}},
+                {"data": {"user": None if not data else {
+                    "lastBroadcast": {
+                        "title": "channel status"
+                    },
+                    "stream": {"game": {
+                        "name": "channel game"
+                    }}
+                }}}
+            ]
         )
-        self.mock.get(
-            "https://api.twitch.tv/kraken/streams/1234",
-            status_code=200 if not failure else 404,
-            json={"stream": None} if not data else {"stream": {
-                "channel": {
-                    "display_name": "channel name",
-                    "status": "channel status",
-                    "game": "channel game"
-                }
-            }}
-        )
-        return self.subject("https://twitch.tv/foo")
 
-    def subject_video(self, data=True, failure=False):
-        self.mock.get(
-            "https://api.twitch.tv/kraken/videos/1337",
-            status_code=200 if not failure else 404,
-            json={} if not data else {
+    def mock_request_video(self, data=True):
+        return self.mock.post(
+            "https://gql.twitch.tv/gql",
+            json={"data": {"video": None if not data else {
                 "title": "video title",
-                "game": "video game",
-                "channel": {
-                    "display_name": "channel name"
+                "game": {
+                    "displayName": "video game"
+                },
+                "owner": {
+                    "displayName": "channel name"
                 }
-            }
+            }}}
         )
-        return self.subject("https://twitch.tv/videos/1337")
 
-    def test_metadata_channel_exists(self):
-        author, title, category = self.subject_channel()
+    def test_metadata_channel(self):
+        mock = self.mock_request_channel()
+        author, title, category = self.subject("https://twitch.tv/foo")
         self.assertEqual(author, "channel name")
         self.assertEqual(title, "channel status")
         self.assertEqual(category, "channel game")
+        self.assertEqual(mock.request_history[0].json(), [
+            {
+                "operationName": "ChannelShell",
+                "extensions": {
+                    "persistedQuery": {
+                        "version": 1,
+                        "sha256Hash": "c3ea5a669ec074a58df5c11ce3c27093fa38534c94286dc14b68a25d5adcbf55"
+                    }
+                },
+                "variables": {
+                    "login": "foo",
+                    "lcpVideosEnabled": False
+                }
+            },
+            {
+                "operationName": "StreamMetadata",
+                "extensions": {
+                    "persistedQuery": {
+                        "version": 1,
+                        "sha256Hash": "059c4653b788f5bdb2f5a2d2a24b0ddc3831a15079001a3d927556a96fb0517f"
+                    }
+                },
+                "variables": {
+                    "channelLogin": "foo"
+                }
+            }
+        ])
 
-    def test_metadata_channel_missing(self):
-        metadata = self.subject_channel(data=False)
-        self.assertEqual(metadata, (None, None, None))
+    def test_metadata_channel_no_data(self):
+        self.mock_request_channel(data=False)
+        author, title, category = self.subject("https://twitch.tv/foo")
+        self.assertEqual(author, None)
+        self.assertEqual(title, None)
+        self.assertEqual(category, None)
 
-    def test_metadata_channel_invalid(self):
-        with self.assertRaises(PluginError):
-            self.subject_channel(failure=True)
-
-    def test_metadata_video_exists(self):
-        author, title, category = self.subject_video()
+    def test_metadata_video(self):
+        mock = self.mock_request_video()
+        author, title, category = self.subject("https://twitch.tv/videos/1337")
         self.assertEqual(author, "channel name")
         self.assertEqual(title, "video title")
         self.assertEqual(category, "video game")
+        self.assertEqual(
+            mock.request_history[0].json(),
+            {
+                "operationName": "VideoMetadata",
+                "extensions": {
+                    "persistedQuery": {
+                        "version": 1,
+                        "sha256Hash": "cb3b1eb2f2d2b2f65b8389ba446ec521d76c3aa44f5424a1b1d235fe21eb4806"
+                    }
+                },
+                "variables": {
+                    "channelLogin": "",
+                    "videoID": "1337"
+                }
+            }
+        )
 
-    def test_metadata_video_missing(self):
-        metadata = self.subject_video(data=False)
-        self.assertEqual(metadata, (None, None, None))
-
-    def test_metadata_video_invalid(self):
-        with self.assertRaises(PluginError):
-            self.subject_video(failure=True)
+    def test_metadata_video_no_data(self):
+        self.mock_request_video(data=False)
+        author, title, category = self.subject("https://twitch.tv/videos/1337")
+        self.assertEqual(author, None)
+        self.assertEqual(title, None)
+        self.assertEqual(category, None)
 
 
 @patch("streamlink.plugins.twitch.log")
 class TestTwitchHosting(unittest.TestCase):
-    def subject(self, channel, hosts=None, disable=False):
+    @staticmethod
+    def subject(channel, hosts=None, disable=False):
         with requests_mock.Mocker() as mock:
-            mock.get(
-                "https://api.twitch.tv/kraken/users?login=foo",
-                json={"users": [{"_id": 1}]}
-            )
+            mock.register_uri(requests_mock.ANY, requests_mock.ANY, exc=requests_mock.exceptions.InvalidRequest)
             if hosts is None:
-                mock.get("https://tmi.twitch.tv/hosts", json={})
+                mock.post("https://gql.twitch.tv/gql", json={})
             else:
-                mock.get(
-                    "https://tmi.twitch.tv/hosts",
-                    [{"json": {
-                        "hosts": [dict(
-                            host_id=host_id,
-                            target_id=target_id,
-                            target_login=target_login,
-                            target_display_name=target_display_name
-                        )]}
-                      } for host_id, target_id, target_login, target_display_name in hosts]
-                )
+                mock.post("https://gql.twitch.tv/gql", response_list=[
+                    {"json": {"data": {"user": {
+                        "id": host[0],
+                        "hosting": None if not host[1:3] else {
+                            "login": host[1],
+                            "displayName": host[2]
+                        }}}}
+                     } for host in hosts
+                ])
 
             session = Streamlink()
             Twitch.bind(session, "tests.plugins.test_twitch")
@@ -378,31 +440,28 @@ class TestTwitchHosting(unittest.TestCase):
             plugin.options.set("disable-hosting", disable)
 
             res = plugin._switch_to_hosted_channel()
-            return res, plugin.channel, plugin._channel_id, plugin.author
+            return res, plugin.channel, plugin.author
 
     def test_hosting_invalid_host_data(self, mock_log):
-        res, channel, channel_id, author = self.subject("foo")
+        res, channel, author = self.subject("foo")
         self.assertFalse(res, "Doesn't stop HLS resolve procedure")
         self.assertEqual(channel, "foo", "Doesn't switch channel")
-        self.assertEqual(channel_id, 1, "Doesn't switch channel id")
         self.assertEqual(author, None, "Doesn't override author metadata")
         self.assertEqual(mock_log.info.mock_calls, [], "Doesn't log anything to info")
         self.assertEqual(mock_log.error.mock_calls, [], "Doesn't log anything to error")
 
     def test_hosting_no_host_data(self, mock_log):
-        res, channel, channel_id, author = self.subject("foo", [(1, None, None, None)])
+        res, channel, author = self.subject("foo", [(1,)])
         self.assertFalse(res, "Doesn't stop HLS resolve procedure")
         self.assertEqual(channel, "foo", "Doesn't switch channel")
-        self.assertEqual(channel_id, 1, "Doesn't switch channel id")
         self.assertEqual(author, None, "Doesn't override author metadata")
         self.assertEqual(mock_log.info.mock_calls, [], "Doesn't log anything to info")
         self.assertEqual(mock_log.error.mock_calls, [], "Doesn't log anything to error")
 
     def test_hosting_host_single(self, mock_log):
-        res, channel, channel_id, author = self.subject("foo", [(1, 2, "bar", "Bar"), (2, None, None, None)])
+        res, channel, author = self.subject("foo", [(1, "bar", "Bar"), (2,)])
         self.assertFalse(res, "Doesn't stop HLS resolve procedure")
         self.assertEqual(channel, "bar", "Switches channel")
-        self.assertEqual(channel_id, 2, "Switches channel id")
         self.assertEqual(author, "Bar", "Overrides author metadata")
         self.assertEqual(mock_log.info.mock_calls, [
             call("foo is hosting bar"),
@@ -411,10 +470,9 @@ class TestTwitchHosting(unittest.TestCase):
         self.assertEqual(mock_log.error.mock_calls, [], "Doesn't log anything to error")
 
     def test_hosting_host_single_disable(self, mock_log):
-        res, channel, channel_id, author = self.subject("foo", [(1, 2, "bar", "Bar")], disable=True)
+        res, channel, author = self.subject("foo", [(1, "bar", "Bar")], disable=True)
         self.assertTrue(res, "Stops HLS resolve procedure")
         self.assertEqual(channel, "foo", "Doesn't switch channel")
-        self.assertEqual(channel_id, 1, "Doesn't switch channel id")
         self.assertEqual(author, None, "Doesn't override author metadata")
         self.assertEqual(mock_log.info.mock_calls, [
             call("foo is hosting bar"),
@@ -423,15 +481,14 @@ class TestTwitchHosting(unittest.TestCase):
         self.assertEqual(mock_log.error.mock_calls, [], "Doesn't log anything to error")
 
     def test_hosting_host_multiple(self, mock_log):
-        res, channel, channel_id, author = self.subject("foo", [
-            (1, 2, "bar", "Bar"),
-            (2, 3, "baz", "Baz"),
-            (3, 4, "qux", "Qux"),
-            (4, None, None, None)
+        res, channel, author = self.subject("foo", [
+            (1, "bar", "Bar"),
+            (2, "baz", "Baz"),
+            (3, "qux", "Qux"),
+            (4,)
         ])
         self.assertFalse(res, "Doesn't stop HLS resolve procedure")
         self.assertEqual(channel, "qux", "Switches channel")
-        self.assertEqual(channel_id, 4, "Switches channel id")
         self.assertEqual(author, "Qux", "Overrides author metadata")
         self.assertEqual(mock_log.info.mock_calls, [
             call("foo is hosting bar"),
@@ -444,14 +501,13 @@ class TestTwitchHosting(unittest.TestCase):
         self.assertEqual(mock_log.error.mock_calls, [], "Doesn't log anything to error")
 
     def test_hosting_host_multiple_loop(self, mock_log):
-        res, channel, channel_id, author = self.subject("foo", [
-            (1, 2, "bar", "Bar"),
-            (2, 3, "baz", "Baz"),
-            (3, 1, "foo", "Foo")
+        res, channel, author = self.subject("foo", [
+            (1, "bar", "Bar"),
+            (2, "baz", "Baz"),
+            (3, "foo", "Foo")
         ])
         self.assertTrue(res, "Stops HLS resolve procedure")
         self.assertEqual(channel, "baz", "Has switched channel")
-        self.assertEqual(channel_id, 3, "Has switched channel id")
         self.assertEqual(author, "Baz", "Has overridden author metadata")
         self.assertEqual(mock_log.info.mock_calls, [
             call("foo is hosting bar"),
